"""Symfem: a symbolic finite element definition library."""

from .create import add_element, create_element, create_reference
from .version import version as __version__

__citation__ = "https://doi.org/10.21105/joss.03556 (Scroggs, 2021)"
__github__ = "https://github.com/mscroggs/symfem"
__git__ = "https://github.com/mscroggs/symfem.git"
